###################################################################################
# Phabricator Task Handler
# Copyright (C) 2016 Luis Araujo <luis.araujo@collabora.co.uk>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  US
###################################################################################

import json
import os.path

from argparse import ArgumentParser
from phabricator import Phabricator
from .version import __version__

PRIORITY_CHOICES=['low', 'normal', 'triage', 'high', 'blocker']

class Cli(object):
    """Command line interface using argparse"""

    def __init__(self):
        self.parser = ArgumentParser(description="phab-handler v{}"
                                     .format(__version__))
        self.parser.add_argument('-c', '--config', metavar='ARCCONFIG',
                                 default=None,
                                 help="Set the arcrc configuration file to use")
        self.parser.add_argument('-t', '--host', default=None,
                                 help="Specify the host to use")
        # Sub-commands
        subparsers = self.parser.add_subparsers()

        # Update Task
        update_parser = subparsers.add_parser('update', help="Update task")
        update_parser.add_argument('id', type=int, help="Task ID")
        update_parser.add_argument('-t', '--title', type=str, help="Update title")
        update_parser.add_argument('-d', '--description', type=str,
                                   help="Update description")
        update_parser.add_argument('-c', '--comment', type=str, help="Add comment")
        update_parser.add_argument('-p', '--priority', choices=PRIORITY_CHOICES,
                                   help="Update priority")
        update_parser.add_argument('-s', '--status', type=str,
                                   help="Update status")
        update_parser.add_argument('--override-status', action='store_true',
                                   help="Ignore status workflow")
        update_parser.set_defaults(func=update)

        # List available statuses
        list_statuses_parse = subparsers.add_parser('list-statuses',
                                                    help="List available statuses")
        list_statuses_parse.set_defaults(func=list_statuses)

        # Who am I?
        whoami_parser = subparsers.add_parser('whoami', help="Show authenticated "
                                              "user name")
        whoami_parser.set_defaults(func=whoami)

    def run(self):
        args = self.parser.parse_args()
        try:
            args.func(args)
        except KeyboardInterrupt:
            pass

class Command(object):

    def __init__(self, args):
        self.args = args
        # Use passed arcrc config file if any, otherwise use the
        # python-phabricator defaults (~/.arcrc).
        if self.args.config:
            api_url, api_token = read_config_file(self.args.config, self.args.host)
            self.phab = Phabricator(host=api_url, token=api_token)
        else:
            self.phab = Phabricator(host=self.args.host)
        # Make sure to always have latest interfaces.
        self.phab.update_interfaces()

    def _statuses(self):
        return self.phab.maniphest.querystatuses()['allStatuses']

class Update(Command):

    def __init__(self, args):
        Command.__init__(self, args)

    def run(self):
        params = { 'id': self.args.id }

        # Update the different values for those passed flags.
        if self.args.title:
            params.update({ 'title': self.args.title })
        if self.args.description:
            params.update({ 'description': self.args.description })
        if self.args.comment:
            params.update({ 'comments': self.args.comment })
        if self.args.priority:
            params.update({ 'priority': { 'low':     50,
                                          'normal':  60,
                                          'triage':  70,
                                          'high':    90,
                                          'blocker': 100 }[self.args.priority] })
        if self.args.status:
            # If the --override-status option is not passed only update status
            # if it moves forward to a higher closed state
            if not self.args.override_status:
                t_status = self.phab.maniphest.info(task_id=self.args.id)['status']
                if self._compare_status(t_status, self.args.status) >= 0:
                    params.update({ 'status': self.args.status })
                else:
                    # Return without doing anything if the status won't be updated.
                    return
            else:
                params.update({ 'status': self.args.status })

        # Only update task if there are values to update.
        if len(params) > 1:
            print self.phab.maniphest.update(**params)

    def _compare_status(self, a ,b):
        # Fetch all available statuses in this phab installation to compare with.
        statuses_order = self._statuses()
        try:
            a_i = statuses_order.index(a)
        except ValueError:
            a_i = 0
        try:
            b_i = statuses_order.index(b)
        except ValueError:
            b_i = 0
        return b_i - a_i

class ListStatuses(Command):

    def __init__(self, args):
        Command.__init__(self, args)

    def run(self):
        for status in self._statuses():
            print status

class WhoAmI(Command):

    def __init__(self, args):
        Command.__init__(self, args)

    def run(self):
        print self.phab.user.whoami()

def update(args):
    Update(args).run()

def list_statuses(args):
    ListStatuses(args).run()

def whoami(args):
    WhoAmI(args).run()

# Configuration files functions.
class ConfigError(Exception):
    pass

def read_config_file(config_file, arg_host=None):
    """This class method is used to fetch the required conduit API
    information from a configuration file."""

    if not os.path.exists(config_file):
        raise ConfigError("File does not exist: {}".format(config_file))

    try:
        with open(config_file, 'r') as conf_data:
            arcrc_config = json.load(conf_data)
    except IOError as err:
        raise ConfigError(err)
    except EnvironmentError as err:
        raise ConfigError(err)

    config = arcrc_config.get('config', None)
    if not config:
        raise ConfigError("No 'config' field found in {}".format(config_file))

    default_host = config.get('default', None)
    if not default_host:
        raise ConfigError("No 'default' host found in {}".format(config_file))

    if 'hosts' not in arcrc_config:
        raise ConfigError("No 'hosts' field found in {}".format(config_file))

    use_host = arg_host if arg_host else default_host
    api_url = api_token = None
    for host in arcrc_config.get('hosts', {}).keys():
        if host.startswith(use_host):
            api_url = host
            api_token = arcrc_config['hosts'][host].get('token', None)

    # Check that api url and token fields were found. 
    if not api_url:
        raise ConfigError('API URL not found in {}'.format(config_file))
    if not api_token:
        raise ConfigError('API Token not found in {}'.format(config_file))

    return (api_url, api_token)
